module Parser where

import Syntax

import Context

import IO

import Monad

-- only for hugs
--isEOF = hugsIsEOF

type Env = Maybe Tag -- Tag of neighbour

parse :: IO Document

parse = liftM head $ ps [[]] [] [emptyContext] [Nothing]

ps :: [Document] -> [Tag] -> [Context] -> [Env] -> IO [Document]
ps docs tags ct env = do 
                     	eof <- isEOF
		     	if eof then return docs
		        	else do 
			               c         <- getChar
			   	       (d,t,ct',env') <- eatChar c docs tags ct env
				       ps d t ct' env'

eatChar :: Char -> [Document] -> [Tag] -> [Context] -> [Env] -> IO ([Document], [Tag], [Context],[Env])

eatChar '<'  docs tags ct env = 
	eatTag docs tags ct env

eatChar ' ' docs@((Ws:r):t) tags ct@(Context _ _ _ _ False _ _ _:_)  env = 
	return (docs,tags,ct,env)
eatChar ' '  (h:t) tags ct@(Context _ _ _ _ False _ _ _:_) (_:er) = 
	return ((Ws:h):t,tags,ct,Nothing:er)
eatChar '\n' docs@((Ws:r):t) tags ct@(Context _ _ _ _ False _ _ _:_)  env = 
	return (docs,tags,ct,env)
eatChar '\n' (h:t) tags ct@(Context _ _ _ _ False _ _ _:_) (_:er) = 
	return ((Ws:h):t,tags,ct,Nothing:er)
eatChar '\t' docs@((Ws:r):t) tags ct@(Context _ _ _ _ False _ _ _:_) env  =
	return (docs,tags,ct,env)
eatChar '\t' (h:t) tags ct@(Context _ _ _ _ False _ _ _:_) (_:er) = 
	return ((Ws:h):t,tags,ct,Nothing:er)
eatChar '\r' docs@((Ws:r):t) tags ct@(Context _ _ _ _ False _ _ _:_) env = 
	return (docs, tags,ct,env)
eatChar '\r' (h:t) tags ct@(Context _ _ _ _ False _ _ _:_) (_:er) = 
	return ((Ws:h):t,tags,ct,Nothing:er)

eatChar c ( ( (Txt l) : hr) : tr) tags ct env = 
	return ( ((Txt (c:l)) : hr) : tr  , tags, ct, env)
eatChar c (h:t) tags ct (_:er) = 
	return (((Txt [c]):h):t,tags,ct,Nothing:er)

eatTag :: [Document] -> [Tag] -> [Context] -> [Env] -> IO ([Document],[Tag],[Context],[Env])
eatTag docs tags ct env =
  do 
     c <- getChar
     restTag
     return (case c of
       		'/' -> popTag docs tags ct env
       		'E' -> pushEM docs tags ct (Nothing:env)
       		'T' -> pushTT docs tags ct (Nothing:env)
       		'P' -> pushPlain docs tags ct (Nothing:env)
       		_   -> pushTag c docs tags ct (Nothing:env))
       
popTag :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
 -- leave out empty tags
popTag ([]:r) (t:ts) ct env = 
	(r,ts,ct,env)
popTag (x:(y:r)) (t:ts) (c':c:cs) (e:f:er) = 
	if c == c' then ((x++y):r,ts,c:cs,e:er)
	           else case (f,y) of 
		   	(Nothing,_) -> (((Tag t x):y):r,ts,c:cs,Just t:er)
			(Just s,Tag _ z:l)  -> if s == t then ((Tag t (x ++ z):l):r,ts,c:cs,f:er)
			                       	         else (((Tag t x):y):r,ts,c:cs,Just t:er)


pushEM :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushEM docs tags (c:cs) env = 
	([]:docs,EM:tags,(enterTag c EM):c:cs,env)
 
pushTT :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushTT docs tags (c:cs) env = 
	([]:docs,TT:tags,(enterTag c TT):c:cs,env)

pushPlain :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushPlain docs tags (c:cs) env = 
	([]:docs,PL:tags,(enterTag c PL):c:cs,env)
 
pushTag :: Char -> [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushTag x docs tags (c:cs) env = 
	let t = getTag x
	in ([]:docs,t:tags,(enterTag c t):c:cs,env)

restTag :: IO ()
restTag = 
  do 
     c <- getChar
     case c of 
       '>' -> return ()
       _   -> restTag
